@@ -1,26 +1,33 @@
+/* -*- mode: C; c-basic-offset: 4; indent-tabs-mode: nil; -*- */
 /* vim:set et sts=4: */
 /* ibus - The Input Bus
- * Copyright (C) 2008-2009 Peng Huang <shawn.p.huang@gmail.com>
- * Copyright (C) 2008-2009 Red Hat, Inc.
+ * Copyright (C) 2008-2010 Peng Huang <shawn.p.huang@gmail.com>
+ * Copyright (C) 2008-2010 Red Hat, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
- * version 2 of the License, or (at your option) any later version.
+ * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
+ * USA
  */
+
+#if !defined (__IBUS_H_INSIDE__) && !defined (IBUS_COMPILATION)
+#error "Only <ibus.h> can be included directly"
+#endif
+
 /**
  * SECTION: ibusobject
  * @short_description: Base object of IBus.
+ * @title: IBusObject
  * @stability: Stable
  *
  * IBusObject is the base object for all objects in IBus.
@@ -28,16 +35,7 @@
 #ifndef __IBUS_OBJECT_H_
 #define __IBUS_OBJECT_H_
 
-/**
- * SECTION: IBusObject
- * @short_description: Base object of IBus.
- *
- *
- * This module provides essential definition and functions for manipulating
- * IBus base object.
- */
 #include <glib-object.h>
-#include "ibusmarshalers.h"
 #include "ibustypes.h"
 #include "ibusdebug.h"
 
@@ -69,12 +67,15 @@ typedef enum {
 #define IBUS_OBJECT_FLAGS(obj)             (IBUS_OBJECT (obj)->flags)
 #define IBUS_OBJECT_SET_FLAGS(obj,flag)    G_STMT_START{ (IBUS_OBJECT_FLAGS (obj) |= (flag)); }G_STMT_END
 #define IBUS_OBJECT_UNSET_FLAGS(obj,flag)  G_STMT_START{ (IBUS_OBJECT_FLAGS (obj) &= ~(flag)); }G_STMT_END
+#define IBUS_OBJECT_IN_DESTRUCTION(obj)    (IBUS_OBJECT_FLAGS (obj) & IBUS_IN_DESTRUCTION)
 #define IBUS_OBJECT_DESTROYED(obj)         (IBUS_OBJECT_FLAGS (obj) & IBUS_DESTROYED)
 
 G_BEGIN_DECLS
 
 typedef struct _IBusObject IBusObject;
 typedef struct _IBusObjectClass IBusObjectClass;
+typedef struct _IBusObjectPrivate IBusObjectPrivate;
+
 /**
  * IBusObject:
  *
@@ -82,9 +83,11 @@ typedef struct _IBusObjectClass IBusObjectClass;
  * private to the #IBusObject and should never be accessed directly.
  */
 struct _IBusObject {
-  GInitiallyUnowned parent;
-  /* instance members */
-  guint32 flags;
+    GInitiallyUnowned parent;
+    /* instance members */
+    guint32 flags;
+
+    IBusObjectPrivate *priv;
 };
 
 typedef void ( *IBusObjectDestroyFunc) (IBusObject *);
